/*	Hardware definitions for 320x240FRS LCD displays
	
		PA.0-7	I/O	LCD D0-7
	
		PE.4	Out	LCD RD_
		PE.0	Out	LCD A0
		PE.3	Out	LCD WR_

		PE.1	Out	LCD CS_

*/

/* LCD control signals */
#define	LCD_RD	4
#define	LCD_A0	0
#define	LCD_WR	3
#define	LCD_CS	1


void MsDelay ( int iDelay )
{
	unsigned long ul0;
	ul0 = MS_TIMER;		// get current timer value
	while ( MS_TIMER < ul0 + (unsigned long) iDelay );
}


/* Low level functions */

#asm
;function requires one parameter:
;@sp+2= data to write
;
LCD_Write::
	ld hl,(sp+2)			; get data value (LSB)
	ld a,l
	ioi ld (PADR),a		; and write it (no shadow, no one else uses port A)
	ld hl,PEDRShadow		; point to control port (shadow)
	ld de,PEDR				; point to control port
	res LCD_WR,(HL)		; Lower WR
	ioi ldd					; do it
	ld hl,PEDRShadow		; point to control port (shadow)
	ld de,PEDR				; point to control port
	set LCD_WR,(HL)		; Rise WR
	ioi ldd					; do it
	ret

#endasm



void LCD_WriteCmd(unsigned char cmd)
{
	BitWrPortI ( PEDR, &PEDRShadow, 1,LCD_A0 );	// Rise A0 (Cmd)
	LCD_Write(cmd);
	BitWrPortI ( PEDR, &PEDRShadow, 0,LCD_A0 );	// Lower A0 (Data)
}

void LCD_WriteStrCmd(unsigned char *cmd,int len)
{
	LCD_WriteCmd(*cmd++);
	while(len--) {
		LCD_Write(*cmd++);
	}
}

unsigned char LCD_ReadData()
{
unsigned char data;
	WrPortI ( SPCR, &SPCRShadow, 0x80 );			// PA0-7 = Inputs, remove data from bus
	BitWrPortI ( PEDR, &PEDRShadow, 1,LCD_A0 );	// Rise A0 (Data)
	BitWrPortI ( PEDR, &PEDRShadow, 0,LCD_RD );	// Lower RD
	data=RdPortI(PADR);									// read data
	BitWrPortI ( PEDR, &PEDRShadow, 1,LCD_RD );	// Rise RD
	BitWrPortI ( PEDR, &PEDRShadow, 0,LCD_A0 );	// Lower A0 (Sts)
	WrPortI ( SPCR, &SPCRShadow, 0x84 );			// PA0-7 = Outputs
	return(data);
}



void LCD_init ()
{
const static unsigned char init_string1[]={
0x40,				// INIT
'\B00110000',	// 32 char CGRAM, 8 pixel chars, no top-screen corr, LCD, normal
0x87,0x7,		// 8x8 characters
39,59,			// 320 pixels, 8 pixel characters => 40 characters per line (43)
239,				// 240 lines (30 txt lines)
40,0				// virtual screen = display screen
};

const static unsigned char init_string2[]={
0x44,				// SCROLL
0, 				// SAD1L: start address screen 1 (0x0000)
0, 				// SAD1H: (for text screen)
239, 				// SL1  : number of scrolling lines screen 1 (239 dec.)
0xB0, 			// SAD2L: start address screen 2 (1200=0x4b0)
0x04, 			// SAD2H: (for the graphics screen)
239, 				// SL2  : number of scrolling lines screen 2
0, 				// SAD3L:
0, 				// SAD3H:
0, 				// SAD4L:
0	 				// SAD4H:
};

const static unsigned char init_string3[]={
0x5D,				// CSRFORM	 
0x04,				// CRX: set the horizontal size of the cursor (4 pixel)
0x86				// CRY: set the vertical size of the cursor (6 pixel) CM : set the cursor display mode to block cursor
};

		
const static unsigned char init_string4[]={
0x5A,				// HDOT_SCR (0x5a), Sets horizontal scroll position (scroll rate).
0					// 1 pixel scroll
};

const static unsigned char init_string5[]={
0x5B,				// OVLAY, Selects layered screen composition and screen text/graphics mode.
'\B00000001'	//MX0, MX1: set the text/graphics overlay mode to EXOR
//		Bit 2-3 : DM1, DM2: set the display mode of screen block 1 and 3
//		Bit 4   : OV: set two or three layer compsition in gfx-mode
//		Bit 5-7 :  0: not used (set to 0)
};

const static unsigned char init_string6[]={
0x59,				// send command DISP_ON (0x59)
'\B00010110'	// cursor flash rate = 2 Hz, SAD1/2 no flashing, SAD3 off
};

	WrPortI ( PEDR,&PEDRShadow,'\B00011010' );		// Make sure CS,RD,WR are HIGH
	WrPortI ( PEDDR,&PEDDRShadow,'\B10011011' );	// PE0,1,4,3,7 = output
	WrPortI ( SPCR, &SPCRShadow, 0x84 );			// PA0-7 = Outputs
	MsDelay ( 1000 );				// wait for LCD to reset itself

	BitWrPortI ( PEDR, &PEDRShadow, 0,LCD_CS );	// Lower CS

	LCD_WriteStrCmd ( init_string1,sizeof(init_string1) );
	LCD_WriteStrCmd ( init_string2,sizeof(init_string2) );
	LCD_WriteStrCmd ( init_string3,sizeof(init_string3) );
	LCD_WriteCmd ( 0x4C );		// CSRDIR_RIGHT, sets the direction of automatic cursor increment to right.
	LCD_WriteStrCmd ( init_string4,sizeof(init_string4) );
	LCD_WriteStrCmd ( init_string5,sizeof(init_string5) );
	LCD_WriteStrCmd ( init_string6,sizeof(init_string6) );
}

/* High level functions */

void LCD_cursor ( int address )
{
	LCD_WriteCmd(0x46);									// CSRW, Sets cursor address.
	LCD_Write(address&0xFF);							// LO byte
	LCD_Write((address>>8)&0xFF);						// HI byte
}

void LCD_fill(unsigned char pattern, int len)
{
int i;
	LCD_WriteCmd(0x42);									// MWRITE
	while(len--)
			LCD_Write(pattern);							// fill with pattern
}

void LCD_cleargfx ( void )
{
	LCD_cursor(1200);							// address gfx screen
	LCD_fill(0,9600);							// clear screen (fill 240 lines of 40 stripes with 0's)
}

void LCD_cleartxt ( void )
{
	LCD_cursor(0);								// address txt screen
	LCD_fill(' ',1200);						// clear screen (fill 30 lines with 40 spaces)
}

void LCD_printat (unsigned int row, unsigned int col, char *ptr)
{
	LCD_cursor (40*row+col);							// set cursor address
	LCD_WriteCmd(0x42);									// MWRITE
 	while (*ptr)
		LCD_Write (*ptr++);
}

/* Plot image data, 8 pixels per byte, MSB left */

void LCD_plot (int x, int y)
{
int strip,bit,data;
	strip=x>>3;											// strip = x/8
	bit=x&'\B0111';									// bit = remainder
	y=1200+strip+40*y;
	LCD_cursor(y);										// address strip in gfx screen
	LCD_WriteCmd(0x43);								// MREAD
	data=LCD_ReadData();								// get data at that address
	data|=(0x80>>bit);								// set requested bit (MSB left)
	LCD_cursor(y);										// address strip in gfx screen again
	LCD_WriteCmd(0x42);								// MWRITE
	LCD_Write(data);
}

/* Read image files, store them as [(long)lenght][data...] in xmem
Image data can be extracted from BMP format (without the header), but
vertically mirrored and color inverted (BMP:0=black;1=white, display: 1=pixel lit=black).
Take a desired image, resize to 320x240, color reduce to 2 colors, mirror verticaly
(rotate 180, flip horizontally), invert colors. Save to file, remove 62 byte header
with your preferred hex editor */

#ximport "cika.pbm" cika
#ximport "rabtip.pbm" rabtip
#ximport "goku.pbm" goku

/* To retrieve image data from xmem, we must trade-off speed and buffer size
define a chunk of data long enough (IMGCHUNK), read from xmem, and send it
to the display, repeat for all chunks (IMGCHUNK*CHUNKS=9600 */

#define IMGCHUNK 240
#define CHUNKS	40

void LCD_dump(long imgdata)
{
int x,y;
unsigned char buffer[IMGCHUNK];

	LCD_cursor(1200);							// address gfx screen
	LCD_WriteCmd(0x42);										// MWRITE
	for(y=0;y<CHUNKS;y++){
		xmem2root(&buffer,imgdata+IMGCHUNK*y+sizeof(long),IMGCHUNK);
																	// read data from xmem (skip lenght)
		for(x=0;x<IMGCHUNK;x++)
			LCD_Write(buffer[x]);							// send data to display
	}
}


/* simple 3D function plotting */



void paraboloide(int PERSP,float SCALE,int TRAZOSX,int TRAZOSY,float X0,float X1,float Y0,float Y1)
{
auto int xx,yy;
auto float x,y,z;
auto float persp,xt,zt;
auto float h1,h2,h3,h4,cs,si;

	persp=PERSP*0.01745329;			// convert to radians
	cs=cos(persp);
	si=sin(persp);
	h1=(X1-X0)/159;
	h2=(Y1-Y0)/(float)(TRAZOSX-1);
	h3=(X1-X0)/(float)(TRAZOSY-1);
	h4=(Y1-Y0)/159;

	for(y=Y0;y<=Y1;y+=h2)
		for(x=X0;x<=X1;x+=h1){
			z=x*x/4+y*y/9-15;
			xt=x-y*cs;
			zt=z-y*si;
			xx=160+(int)(SCALE*xt);
			yy=120-(int)(SCALE*zt);
			LCD_plot(xx,yy);
		}
	for(x=X0;x<=X1;x+=h3)
		for(y=Y0;y<=Y1;y+=h4) {
			z=x*x/4+y*y/9-15;
			xt=x-y*cs;
			zt=z-y*si;
			xx=160+(int)(SCALE*xt);
			yy=120-(int)(SCALE*zt);
			LCD_plot(xx,yy);
		}
}

void elipsoide(int PERSP,float SCALE,int TRAZOSX,int TRAZOSY,float X0,float X1,float Y0,float Y1)
{
auto int xx,yy;
auto float x,y,z;
auto float persp,xt,yt,zt;
auto float h1,h2,h3,h4,cs,si;

	persp=PERSP*0.01745329;			// convert to radians
	cs=cos(persp);
	si=sin(persp);
	h1=(X1-X0)/159;
	h2=(Y1-Y0)/(float)(TRAZOSX-1);
	h3=(X1-X0)/(float)(TRAZOSY-1);
	h4=(Y1-Y0)/159;

	for(y=Y0;y<=Y1;y+=h2)
		for(x=X0;x<=X1;x+=h1){
			xt=cos(x)*cos(y);
			yt=sin(x)*cos(y);
			zt=sin(y);
			xt-=yt*cs;
			zt-=yt*si;
			xx=160+(int)(SCALE*xt);
			yy=120-(int)(SCALE*zt);
			LCD_plot(xx,yy);
		}
	for(x=X0;x<=X1;x+=h3)
		for(y=Y0;y<=Y1;y+=h4) {
			xt=cos(x)*cos(y);
			yt=sin(x)*cos(y);
			zt=sin(y);
			xt-=yt*cs;
			zt-=yt*si;
			xx=160+(int)(SCALE*xt);
			yy=120-(int)(SCALE*zt);
			LCD_plot(xx,yy);
		}
}

/* MAIN PROGRAM */

main()
{
int i;

	LCD_init();
	while(1){

		LCD_cleartxt();
		LCD_cleargfx();
		MsDelay ( 1000 );						// wait 1 sec
		LCD_cursor(1200);						// address gfx screen
		LCD_fill(0xFF,9600);					// black fill
		MsDelay ( 3000 );						// wait 3 secs
		LCD_cleargfx();
		LCD_printat(0,0,"Graficos en 3D en tiempo real:");
		paraboloide(30,6,20,20,-8.0,8.0,-10.0,10.0);
		MsDelay ( 10000 );					// wait 10 secs
		LCD_cleartxt();						// remove prior text
		LCD_dump(cika);						// show gfx
		MsDelay ( 3000 );						// wait 3 secs
		LCD_cleargfx();
		LCD_printat(0,0,"Graficos en 3D en tiempo real:");
		elipsoide(45,80.0,30,30,0,6.28,-3.14,3.14);
		MsDelay ( 10000 );					// wait 10 secs

		LCD_cleartxt();						// remove prior text
		LCD_dump(goku);						// show gfx
		LCD_printat(0,0,"Texto y graficos");
		LCD_printat(1,2,"superpuestos");
		MsDelay ( 6000 );						// wait 3 secs
		LCD_cleartxt();						// remove prior text
		LCD_dump(rabtip);						// show gfx
		LCD_printat(25,8,"Cika Electronica S.R.L.");
		LCD_printat(29,30,"julio 2003");
		MsDelay ( 10000 );					// wait 10 secs
	}
}
